(function(module) {
	module.factory('ws', function ($q, $rootScope) {
			var ws;
			var server_version;
			var state = 'idle';
			var next_message_id = 1;
			var requests = [];
			var callbacks = {};
			var listeners = {};
			var ping_timer;
			function handle_response(message) {
				const json = JSON.parse(message.data);
				console.log(json);
				if (json.event) {
					$rootScope.$broadcast('things_happened', { data: json });
					return emit(json.event, json);
				} else if (json.version) {
					if (server_version && server_version != json.version) {
						console.log("Version mismatch, reloading");
						window.location.reload();
					}
					server_version = json.version;
				} else {
					const key = 'r'+json.message_id;
					if (callbacks.hasOwnProperty(key)) {
						var promise = callbacks[key];
						$rootScope.$evalAsync(function() {
							if (json.status == "success" || json.status == "pending") {
								promise.resolve(json);
							} else {
								promise.reject(json);
							}
						});
						delete callbacks[key];
					}
				}
			}
			function connect() {
				if (state == 'idle') {
					console.log("Connecting...");
					state = 'connecting';
					var proto = location.protocol == "http" ? "ws" : "wss";
					ws = new WebSocket(proto + '://' + location.host + '/');
					ws.onclose = function () {
						clearInterval(ping_timer);
						ws = null;
						requests = [];
						callbacks = {};
						state = 'idle';
						if (state == 'connected') {
							emit('disconnect');
						}
						setTimeout(connect, 3000);
					};
					ws.onerror = function() {
						clearInterval(ping_timer);
						var prev_state = state;
						ws = null;
						requests = [];
						callbacks = {};
						state = 'idle';
						if (prev_state == 'connected') {
							emit('disconnect');
						}
						setTimeout(connect, 3000);
					};
					ws.onmessage = handle_response;
					ws.onopen = function() {
						state = 'connected';
						flush_requests();
						ping_timer = setInterval(function() {
							send_request({action: 'ping'});
						}, 180000);
						emit('connect');
					};
				}
			}
			function flush_requests() {
				while (state == 'connected' && requests.length) {
					var request = requests.shift();
					console.log(JSON.parse(request));
					ws.send(request);
				}
			}
			function send_request(request) {
				var defer = $q.defer();
				request.message_id = next_message_id++;
				try {
					requests.push(JSON.stringify(request));
					callbacks['r'+request.message_id] = defer;
					flush_requests();
				} catch (e) {
					console.log(e);
				}
				return defer.promise;
			}
			function emit(event) {
				var args = [].slice.call(arguments, 1);
				$rootScope.$evalAsync(function() {
					var callbacks = (listeners['$' + event] || []).slice(0);
					for (var i = 0; i < callbacks.length; ++i) {
						callbacks[i].apply(this, args);
					}
				});
			}
			function add_listener(event, callback, scope) {
				(listeners['$' + event] = listeners['$' + event] || []).push(callback);
				if (scope) {
					scope.$on('$destroy', function() {
						remove_listener(event, callback);
					});
				}
				return callback;
			}
			function remove_listener(event, callback) {
				if (arguments.length == 0) {
					listeners = {};
				} else if (arguments.length == 1) {
					delete listeners['$' + event];
				} else {
					var callbacks = listeners['$' + event];
					if (callbacks != undefined) {
						for (var i = 0; i < callbacks.length; ++i) {
							if (callbacks[i] === callback) {
								callbacks.splice(i, 1);
								break;
							}
						}
						if (callbacks.length == 0) {
							delete listeners['$' + event];
						}
					}
				}
			}
			connect();
			return {
				reconnect: function() {
					if (ws && ws.readyState < 2) {
						ws.close();
					}
				},
				on: function() {
					add_listener.apply(null, arguments);
					return this;
				},
				off: function() {
					remove_listener.apply(null, arguments);
					return this;
				},
				emit: function() {
					// NOTE: unused, untested
					emit.apply(null, arguments);
					return this;
				},
				cancel: function(task_id) {
					return send_request({action: 'cancel', task_id: task_id});
				},
				apps: function() {
					return send_request({action: 'apps'});
				},
				authenticate: function(session_id) {
					return send_request({action: 'authenticate', session_id: session_id});
				},
				containers: function() {
					return send_request({action: 'containers'});
				},
				container: function(container_uuid) {
					return send_request({action: 'container', uuid: container_uuid});
				},
				start_container: function(container_uuid) {
					return send_request({action: 'start-container', uuid: container_uuid});
				},
				stop_container: function(container_uuid) {
					return send_request({action: 'stop-container', uuid: container_uuid});
				},
				create_container: function(container, whitelist, services, apps) {
					container.memory = parseFloat(container.memory);
					container.cpu = parseFloat(container.cpu);
					container.diskspace = parseFloat(container.diskspace);
					angular.forEach(services, function(service) {
						service.internal_port = parseInt(service.internal_port);
						service.external_port = parseInt(service.external_port);
					});
					return send_request({
						action: "create-container",
						item: { ...container, image: container.image.alias },
						whitelist: whitelist,
						services: services,
						apps: apps.filter((app) => app.status || [ "extensions", "tmx-basic", "tmx-advanced" ].includes(app.name))
							.reduce((obj, app) => { obj[app.name] = ((app.max ?? "") === "") ? -1 : (parseInt(app.max) || 0); return obj; }, {}),
					});
				},
				remove_container: function(container_uuid) {
					return send_request({action: 'remove-container', uuid: container_uuid || null});
				},
				update_container: function(container, whitelist, services, apps, did_routing) {
					container.memory = parseFloat(container.memory);
					container.cpu = parseFloat(container.cpu);
					container.diskspace = parseFloat(container.diskspace);
					angular.forEach(services, function(service) {
						service.internal_port = parseInt(service.internal_port);
						service.external_port = parseInt(service.external_port);
					});
					console.log(apps);
					return send_request({
						action: "update-container",
						item: container,
						whitelist: whitelist,
						services: services,
						apps: apps.filter((app) => app.status || [ "extensions", "tmx-basic", "tmx-advanced" ].includes(app.name))
							.reduce((obj, app) => { obj[app.name] = ((app.max ?? "") === "") ? -1 : (parseInt(app.max) || 0); return obj; }, {}),
						did_routes: did_routing,
					});
				},
				events: function(success, error) {
					return send_request({action: 'events', amount: 100});
				},
				resellers: function() {
					return send_request({action: 'resellers'});
				},
				reseller: function(reseller_uuid) {
					return send_request({action: 'reseller', uuid: reseller_uuid || null});
				},
				create_reseller: function(reseller) {
					return send_request({action: 'create-reseller', item: reseller});
				},
				remove_reseller: function(reseller_uuid) {
					return send_request({action: 'remove-reseller', uuid: reseller_uuid || null});
				},
				update_reseller: function(reseller) {
					reseller.uuid = reseller.uuid || null;
					return send_request({action: 'update-reseller', item: reseller});
				},
				resource_limits: function() {
					return send_request({action: 'resource-limits'});
				},
				update_resource_limits: function(resource_limits) {
					plans = angular.copy(resource_limits);
					angular.forEach(plans, function(plan) {
						plan.memory = parseFloat(plan.memory); // convert ot MB
						plan.cpu = parseFloat(plan.cpu);
						plan.diskspace = parseFloat(plan.diskspace);
					});
					return send_request({action: 'update-resource-limits', items: plans});
				},
				system_log: function() {
					return send_request({action: 'events', type: "error", amount: 100});
				},
				login: function(username, password) {
					return send_request({action: 'authenticate', username: username, password: password});
				},
				logout: function(session_id) {
					return send_request({action: 'logout', session_id: session_id});
				},
				search: function(query) {
					return send_request({action: 'search', query: query});
				},
				settings: function() {
					return send_request({action: 'settings'});
				},
				tasks: function() {
					return send_request({action: 'tasks'});
				},
				update_settings: function(settings, whitelist) {
					settings.cpu_overprovision = parseFloat(settings.cpu_overprovision);
					settings.memory_overprovision = parseFloat(settings.memory_overprovision);
					return send_request({
						action: 'update-settings',
						settings: settings,
						whitelist: whitelist
					});
				},
				images: function() {
					return send_request({action: 'images'});
				},
				interfaces: function() {
					return send_request({action: 'interfaces'});
				},
				profile: function() {
					return send_request({action: 'profile' });
				},
				update_profile: function(settings) {
					return send_request({action: 'update-profile', item: settings});
				},
				version: function() {
					return send_request({action: 'version'});
				},
				generate_report: function(settings) {
					return send_request({action: 'gen-report', item: settings});
				},
				start_containers: function(uuids) {
					return send_request({action: 'start-containers', uuids: uuids});
				},
				stop_containers: function(uuids) {
					return send_request({action: 'stop-containers', uuids: uuids});
				},
				upgrade_containers: function(uuids) {
					return send_request({action: 'upgrade-containers', uuids: uuids});
				},
				validate_password: function(password) {
					return send_request({action: 'validate-password', 'password': password});
				},
				check_strength: function(string) {
					return send_request({action: 'check-strength', 'string': string});
				},
				host_whitelist: function() {
					return send_request({action: 'host-whitelist'});
				},
				confirm_changes: function(token) {
					return send_request({action: 'confirm-changes', token: token});
				},
				'set': function(name, value) {
					return send_request({action: 'set', values: {[name]: value}});
				},
				start_dump: function() {
					return send_request({action: 'start-dump'});
				},
				stop_dump: function(token) {
					return send_request({action: 'stop-dump', 'uuid':token});
				},
				whitelist_services: function() {
					return send_request({action: 'whitelist-services'});
				},
				update_whitelist_services: function(services) {
					angular.forEach(services, function(service) {
						service.internal_port = parseInt(service.internal_port);
					});
					return send_request({action: 'update-whitelist-services', services: services});
				},
				did_routing: function() {
					return send_request({action: 'did-routing'});
				},
				container_did_routing: function(uuid) {
					return send_request({action: 'container-did-routing', 'uuid': uuid});
				},
				update_didrouting: function(did_routing, status) {
					return send_request({
						action: 'update-did-routing',
						'did_routes': did_routing,
						'did_routes_enabled': status
					})
				},
				export_didrouting: function(type) {
					return send_request({action: 'export-did-routing', 'type': type})
				},
				import_didrouting: function(file) {
					return send_request({action: 'import-did-routing', 'file': file})
				},
			};
		}
	);
	module.factory('popup', function ($rootScope) {
		return {
			show: function(message) {
				$rootScope.$broadcast('eventFired', { data: message });
			}
		};
	});
	module.factory('translation_loader', function ($q, $http) {
		return function (options) {
			var defer = $q.defer();
			$http({method: 'GET', url: `/l10n/${options.key}.json?${Date.now()}`}).then(function (response) {
				defer.resolve(response.data);
			});
			return defer.promise;
		};
	});
	module.factory('reseller_srv', function($mdDialog, $translate) {
		return {
			confirm: function(data) {
				var confirm = $mdDialog.confirm()
					.theme('warning')
					.title($translate.instant('global.warning'))
					.htmlContent($translate.instant('reseller.remove_message', data))
					.ok($translate.instant('reseller.confirm'))
					.cancel($translate.instant('reseller.cancel'));
				return confirm;
			}
		}
	});
	module.factory('message', function($mdDialog, $translate) {
		return {
			confirm: function(data) {
				var confirm = $mdDialog.confirm()
					.theme('warning')
					.title($translate.instant('global.warning'))
					.htmlContent(data.content)
					.ok($translate.instant('reseller.confirm'))
					.cancel($translate.instant('reseller.cancel'));
				return $mdDialog.show(confirm);
			},
			confirm_1: function(data) {
				return $mdDialog.show({
					parent: angular.element(document.body),
					clickOutsideToClose:false,
					escapeToClose:false,
					template: `
					<md-dialog aria-label="List dialog" class="{{theme}}">
						<md-toolbar>
							<div class="md-toolbar-tools">
								<h2>{{title}}</h2>
							</div>
						</md-toolbar>
						<md-dialog-content class="app_prompt">
							<div class="md-dialog-content">
								<p ng-bind-html=content></p>
							</div>
						</md-dialog-content>
						<md-dialog-actions>
							<md-button ng-click="closeDialog()" class="md-warn" md-autofocus>{{cancel}}</md-button>
							<md-button ng-click="confirmDialog()" class="md-primary">{{confirm}}</md-button>
						</md-dialog-actions>
					</md-dialog>`,
					controller: function ($scope) {
						$scope.title = $translate.instant(data.title);
						$scope.confirm = $translate.instant(data.confirm);
						$scope.cancel = $translate.instant(data.cancel);
						$scope.content = $translate.instant(data.content);
						$scope.theme = data.theme;
						$scope.closeDialog = function() {
							$mdDialog.hide(false);
						}
						$scope.confirmDialog = function() {
							$mdDialog.hide(true);
						}
					}
				})
			},
			error: function(errors) { //accepts assoc array of error strings
				var errors = Object.keys(errors).map(function(key){
					return errors[key];
				}).join("<br/>");

				$mdDialog.show(
					$mdDialog.alert()
						.theme('error')
						.parent(angular.element(document.querySelector('#popupContainer')))
						.clickOutsideToClose(true)
						.title($translate.instant('global.error'))
						.htmlContent(errors)
						.clickOutsideToClose(true)
						.ok($translate.instant('global.ok'))
				);
			},
			password: function(data) {
				return $mdDialog.show({
					parent: angular.element(document.body),
					clickOutsideToClose:false,
					escapeToClose:false,
					template: `
					<md-dialog aria-label="List dialog" class="{{theme}}">
						<md-toolbar>
							<div class="md-toolbar-tools">
								<h2>{{title}}</h2>
							</div>
						</md-toolbar>
						<md-dialog-content class="app_prompt">
							<div class="md-dialog-content">
								<p ng-bind-html=content></p>
							</div>
							<form>
								<input type="password" md-autofocus ng-model="result" ng-keypress="($event.charCode==13) && result.length > 0 ? confirmDialog() : ''"/>
							</form>
						</md-dialog-content>
						<md-dialog-actions>
							<md-button ng-click="closeDialog()" class="md-warn">{{cancel}}</md-button>
							<md-button ng-click="confirmDialog()" class="md-primary" ng-disabled="!result">{{confirm}}</md-button>
						</md-dialog-actions>
					</md-dialog>`,
					controller: function ($scope) {
						$scope.title = $translate.instant(data.title);
						$scope.confirm = $translate.instant(data.confirm);
						$scope.cancel = $translate.instant(data.cancel);
						$scope.content = $translate.instant(data.content);
						$scope.theme = data.theme;
						$scope.closeDialog = function() {
							$mdDialog.hide();
						}
						$scope.confirmDialog = function() {
							$mdDialog.hide(this.result);
						}
					}
				})
			},
			prompt: function(data) {
				var propmt = $mdDialog.prompt()
				.theme('prompt')
				.title($translate.instant(data.title))
				.htmlContent($translate.instant(data.content))
				.cancel($translate.instant('tenant.cancel'))
				.ok($translate.instant('tenant.confirm'))
				.required(true);
				return $mdDialog.show(propmt);
			},
			prompt_1: function(data) {
				console.log(data);
				var parentEl = angular.element(document.body);
				return $mdDialog.show({
					parent: parentEl,
					clickOutsideToClose:false,
					escapeToClose:false,
					template: `
					<md-dialog aria-label="List dialog">
						<md-toolbar>
							<div class="md-toolbar-tools">
								<h2>{{title}}</h2>
							</div>
						</md-toolbar>
						<md-dialog-content class="app_prompt">
							<div class="md-dialog-content">
								<p ng-bind-html=activate_warning></p>
								<p ng-bind-html="popup"></p>
							</div>
						</md-dialog-content>
						<md-dialog-actions>
							<md-button ng-click="closeDialog()" class="md-warn" md-autofocus>{{cancel}}</md-button>
							<md-button ng-click="addApp()" class="md-primary">{{agree}}</md-button>
						</md-dialog-actions>
					</md-dialog>`,
					controller: function DialogController($scope, $mdDialog) {
						$scope.app = data.content;
						$scope.app.toggle_status = false;
						$scope.title = data.title;
						$scope.cancel = $translate.instant('tenant.cancel');
						$scope.agree = $translate.instant('tenant.agree');
						$scope.popup = data.popup_text;
						$scope.activate_warning = data.warning;
						$scope.closeDialog = function() {
							$mdDialog.hide();
						}
						$scope.addApp = function() {
							$scope.app.toggle_status = true;
							$mdDialog.hide();
						}
					}
				})
			},
			prompt_2: function() {
				return $mdDialog.show({
					parent: angular.element(document.body),
					clickOutsideToClose:false,
					escapeToClose:false,
					template: `
					<md-dialog aria-label="List dialog">
						<md-toolbar>
							<div class="md-toolbar-tools">
								<h2 translate>global.notice</h2>
							</div>
						</md-toolbar>
						<md-dialog-content class="app_prompt">
							<div class="md-dialog-content">
								<p translate>global.system_whitelist_unconfigured</p>
								<p ng-bind-html="popup"></p>
							</div>
						</md-dialog-content>
						<md-dialog-actions>
							<label><input type="checkbox" ng-model="acknowledged" ng-change="acknowledge()"> I understand the risk</label>
							<md-button ng-click="dismiss()" class="md-warn" ng-disabled="!acknowledged">{{dismiss_label}}</md-button>
							<md-button ng-click="close()" class="md-primary">{{close_label}}</md-button>
						</md-dialog-actions>
					</md-dialog>`,
					controller: function DialogController($scope, $mdDialog, ws) {
						// md-button breaks translate somehow
						$scope.close_label = $translate.instant("global.close");
						$scope.dismiss_label = $translate.instant("global.dismiss");
						$scope.close = function() {
							$mdDialog.hide();
						};
						$scope.dismiss = function() {
							ws.set('system_whitelist_notification_dismissed', true).then(function() {
								$mdDialog.hide();
							});
						};
					}
				})
			},
			test_config: function() {
				return $mdDialog.show({
					parent: angular.element(document.body),
					clickOutsideToClose:false,
					escapeToClose:false,
					template: `
					<md-dialog>
						<md-dialog-content class="app_prompt">
							<div class="md-dialog-content">
								<p translate>test_config.content</p>
							</div>
						</md-dialog-content>
					</md-dialog>`,
					controller: function () {}
				})
			},
			info: function(app) {
				var parentEl = angular.element(document.body);
				return $mdDialog.show({
					parent: parentEl,
					clickOutsideToClose:true,
					template: `
					<md-dialog aria-label="List dialog">
						<md-toolbar>
							<div class="md-toolbar-tools">
								<h2>{{displayname}}</h2>
								<span flex></span>
								<md-button class="md-icon-button" ng-click="close()">
									<md-icon class="fa fa-close"></<md-icon>
								</md-button>
							</div>
						</md-toolbar>
						<md-dialog-content class="app_prompt">
							<div class="md-dialog-content">
								<h2></h2>
								<p ng-bind-html="details"></p>
							</div>
						</md-dialog-content>
					</md-dialog>`,
					controller: function DialogController($scope, $mdDialog) {
						$scope.displayname = app.content.displayname;
						$scope.details = $translate.instant(`app.${app.content.name}.details`);
						$scope.cancel = $translate.instant('tenant.close');
						$scope.close = function() {
							$mdDialog.hide();
						}
					}
				})
			},
		}
	});
	module.factory('multiselect', function($translate) {
		return {
			localization: function() {
				return {
					selectAll       : $translate.instant('global.select_all'),
					selectNone      : $translate.instant('global.select_none'),
					search          : $translate.instant('global.search'),
					nothingSelected : $translate.instant('global.no_selected'),
					reset           : $translate.instant('global.reset')
				}
			}
		}
	});
	module.factory('toaster', function(toastr, $translate) {
		return {
			error: function(message) {
				toastr.error(message, $translate.instant('settings.error'), {
					'positionClass' : 'toast-top-center',
					'timeOut': 10000
				});
			},
			success: function(message) {
				toastr.success(message);
			},
			success_centered: function(message) {
				toastr.success(message, $translate.instant('global.success'), {
					'positionClass' : 'toast-top-center',
					'timeOut': 10000
				});
			},
			warning: function(message) {
				toastr.warning(message, $translate.instant('global.warning'), {
					'positionClass' : 'toast-top-center',
					'timeOut': 10000
				});
			}
		}
	});

	module.factory('about_dialog', function($rootScope) {
		return {
			show: function() { $rootScope.about_visible = true; },
			hide: function() { $rootScope.about_visible = false; }
		}
	});

	module.service('timer_clock', function($interval) {
		var counter;
		var ref;
		return {
			set: function(seconds) {
				counter = seconds;
			},
			start: function() {
				counter = 10 * 60;
				var tik = function () {
					counter--;
					if(counter == 0) {
						$interval.cancel(ref);
					}
				}
				ref = $interval(tik, 1000);
				return ref;
			},
			get: function() {
				return counter;
			},
			stop: function() {
				$interval.cancel(ref);
			}
		};
	});

})(angular.module('cpbx', ['ng']));
